"""Transactional event publishing aspect for automatic domain event publishing."""

from inspect import iscoroutinefunction
from typing import Any

from spakky.core.aop.aspect import Aspect, AsyncAspect
from spakky.core.aop.interfaces.aspect import IAspect, IAsyncAspect
from spakky.core.aop.pointcut import After, AfterReturning
from spakky.core.pod.annotations.order import Order
from spakky.data.aspects.transactional import Transactional
from spakky.data.persistency.aggregate_collector import AggregateCollector

from spakky.event.event_publisher import (
    IAsyncEventPublisher,
    IEventPublisher,
)


@Order(1)
@AsyncAspect()
class AsyncTransactionalEventPublishingAspect(IAsyncAspect):
    _collector: AggregateCollector
    _publisher: IAsyncEventPublisher

    def __init__(
        self,
        collector: AggregateCollector,
        publisher: IAsyncEventPublisher,
    ) -> None:
        self._collector = collector
        self._publisher = publisher

    @AfterReturning(lambda x: Transactional.exists(x) and iscoroutinefunction(x))
    async def after_returning_async(self, result: Any) -> None:
        for aggregate in self._collector.all():
            for event in aggregate.events:
                await self._publisher.publish(event)
            aggregate.clear_events()

    @After(lambda x: Transactional.exists(x) and iscoroutinefunction(x))
    async def after_async(self) -> None:
        self._collector.clear()


@Order(1)
@Aspect()
class TransactionalEventPublishingAspect(IAspect):
    _collector: AggregateCollector
    _publisher: IEventPublisher

    def __init__(
        self,
        collector: AggregateCollector,
        publisher: IEventPublisher,
    ) -> None:
        self._collector = collector
        self._publisher = publisher

    @AfterReturning(lambda x: Transactional.exists(x) and not iscoroutinefunction(x))
    def after_returning(self, result: Any) -> None:
        for aggregate in self._collector.all():
            for event in aggregate.events:
                self._publisher.publish(event)
            aggregate.clear_events()

    @After(lambda x: Transactional.exists(x) and not iscoroutinefunction(x))
    def after(self) -> None:
        self._collector.clear()



"""Spakky Event Aspects package.

This package provides aspects for event-driven architecture:
- TransactionalEventPublishingAspect: Automatic domain event publishing after transactions
"""

from spakky.event.aspects.transactional_event_publishing import (
    AsyncTransactionalEventPublishingAspect,
    TransactionalEventPublishingAspect,
)

__all__ = [
    "AsyncTransactionalEventPublishingAspect",
    "TransactionalEventPublishingAspect",
]



from spakky.core.application.application import SpakkyApplication

from spakky.event.aspects.transactional_event_publishing import (
    AsyncTransactionalEventPublishingAspect,
    TransactionalEventPublishingAspect,
)
from spakky.event.bus.transport_event_bus import (
    AsyncDirectEventBus,
    DirectEventBus,
)
from spakky.event.mediator.domain_event_mediator import (
    AsyncEventMediator,
    EventMediator,
)
from spakky.event.post_processor import EventHandlerRegistrationPostProcessor
from spakky.event.publisher.domain_event_publisher import (
    AsyncEventPublisher,
    EventPublisher,
)


def initialize(app: SpakkyApplication) -> None:
    app.add(AsyncTransactionalEventPublishingAspect)
    app.add(TransactionalEventPublishingAspect)

    app.add(EventMediator)
    app.add(AsyncEventMediator)
    app.add(EventPublisher)
    app.add(AsyncEventPublisher)
    app.add(DirectEventBus)
    app.add(AsyncDirectEventBus)

    app.add(EventHandlerRegistrationPostProcessor)



"""Spakky Event package - Event-driven architecture support."""

from spakky.event.aspects import (
    AsyncTransactionalEventPublishingAspect,
    TransactionalEventPublishingAspect,
)
from spakky.event.bus import (
    AsyncDirectEventBus,
    DirectEventBus,
)
from spakky.event.error import (
    AbstractSpakkyEventError,
    DuplicateEventHandlerError,
    InvalidMessageError,
)
from spakky.event.event_consumer import (
    IAsyncEventConsumer,
    IEventConsumer,
)
from spakky.event.event_dispatcher import (
    IAsyncEventDispatcher,
    IEventDispatcher,
)
from spakky.event.event_publisher import (
    IAsyncEventBus,
    IAsyncEventPublisher,
    IAsyncEventTransport,
    IEventBus,
    IEventPublisher,
    IEventTransport,
)
from spakky.event.mediator import (
    AsyncEventMediator,
    EventMediator,
)
from spakky.event.post_processor import EventHandlerRegistrationPostProcessor
from spakky.event.publisher import (
    AsyncEventPublisher,
    EventPublisher,
)

__all__ = [
    # Publisher Interfaces
    "IAsyncEventPublisher",
    "IEventPublisher",
    # Bus Interfaces
    "IAsyncEventBus",
    "IEventBus",
    # Transport Interfaces
    "IAsyncEventTransport",
    "IEventTransport",
    # Consumer Interfaces
    "IAsyncEventConsumer",
    "IEventConsumer",
    # Dispatcher Interfaces
    "IAsyncEventDispatcher",
    "IEventDispatcher",
    # Mediator Implementations
    "AsyncEventMediator",
    "EventMediator",
    # Publisher Implementations
    "AsyncEventPublisher",
    "EventPublisher",
    # Bus Implementations
    "AsyncDirectEventBus",
    "DirectEventBus",
    # Post-Processors
    "EventHandlerRegistrationPostProcessor",
    # Aspects
    "AsyncTransactionalEventPublishingAspect",
    "TransactionalEventPublishingAspect",
    # Errors
    "AbstractSpakkyEventError",
    "DuplicateEventHandlerError",
    "InvalidMessageError",
]

from spakky.core.application.plugin import Plugin

PLUGIN_NAME = Plugin(name="spakky-event")
"""Plugin identifier for the Spakky Event package."""



from abc import ABC, abstractmethod
from typing import Awaitable, Callable, TypeAlias, TypeVar

from spakky.domain.models.event import AbstractEvent

EventT_contra = TypeVar("EventT_contra", bound=AbstractEvent, contravariant=True)
EventHandlerCallback: TypeAlias = Callable[[EventT_contra], None]
AsyncEventHandlerCallback: TypeAlias = Callable[[EventT_contra], Awaitable[None]]


class IEventConsumer(ABC):
    @abstractmethod
    def register(
        self,
        event: type[EventT_contra],
        handler: EventHandlerCallback[EventT_contra],
    ) -> None: ...


class IAsyncEventConsumer(ABC):
    @abstractmethod
    def register(
        self,
        event: type[EventT_contra],
        handler: AsyncEventHandlerCallback[EventT_contra],
    ) -> None: ...



"""Event publisher implementations.

This module provides event publishers that route events by type:
- AbstractDomainEvent → EventMediator (in-process dispatch)
- AbstractIntegrationEvent → IEventBus (external transport)
"""

from spakky.core.pod.annotations.pod import Pod
from spakky.domain.models.event import (
    AbstractDomainEvent,
    AbstractEvent,
    AbstractIntegrationEvent,
)

from spakky.event.event_dispatcher import (
    IAsyncEventDispatcher,
    IEventDispatcher,
)
from spakky.event.event_publisher import (
    IAsyncEventBus,
    IAsyncEventPublisher,
    IEventBus,
    IEventPublisher,
)


@Pod()
class EventPublisher(IEventPublisher):
    _dispatcher: IEventDispatcher
    _bus: IEventBus

    def __init__(
        self,
        dispatcher: IEventDispatcher,
        bus: IEventBus,
    ) -> None:
        self._dispatcher = dispatcher
        self._bus = bus

    def publish(self, event: AbstractEvent) -> None:
        match event:
            case AbstractDomainEvent():
                self._dispatcher.dispatch(event)
            case AbstractIntegrationEvent():
                self._bus.send(event)
            case _:  # pragma: no cover - 방어적 분기 (정상 흐름 불가)
                raise AssertionError(f"Unknown event type: {type(event)!r}")


@Pod()
class AsyncEventPublisher(IAsyncEventPublisher):
    _dispatcher: IAsyncEventDispatcher
    _bus: IAsyncEventBus

    def __init__(
        self,
        dispatcher: IAsyncEventDispatcher,
        bus: IAsyncEventBus,
    ) -> None:
        self._dispatcher = dispatcher
        self._bus = bus

    async def publish(self, event: AbstractEvent) -> None:
        match event:
            case AbstractDomainEvent():
                await self._dispatcher.dispatch(event)
            case AbstractIntegrationEvent():
                await self._bus.send(event)
            case _:  # pragma: no cover - 방어적 분기 (정상 흐름 불가)
                raise AssertionError(f"Unknown event type: {type(event)!r}")



"""Spakky Event Publisher package.

This package provides event publisher implementations with type-based routing.
"""

from spakky.event.publisher.domain_event_publisher import (
    AsyncEventPublisher,
    EventPublisher,
)

__all__ = [
    "AsyncEventPublisher",
    "EventPublisher",
]



"""Default EventBus implementations that delegate to EventTransport."""

from pydantic import TypeAdapter
from spakky.core.pod.annotations.pod import Pod
from spakky.domain.models.event import AbstractIntegrationEvent

from spakky.event.event_publisher import (
    IAsyncEventBus,
    IAsyncEventTransport,
    IEventBus,
    IEventTransport,
)


@Pod()
class DirectEventBus(IEventBus):
    _transport: IEventTransport
    _adapters: dict[type, TypeAdapter[AbstractIntegrationEvent]]

    def __init__(self, transport: IEventTransport) -> None:
        self._transport = transport
        self._adapters = {}

    def send(self, event: AbstractIntegrationEvent) -> None:
        event_type = type(event)
        if event_type not in self._adapters:
            self._adapters[event_type] = TypeAdapter(event_type)
        adapter = self._adapters[event_type]
        self._transport.send(event.event_name, adapter.dump_json(event))


@Pod()
class AsyncDirectEventBus(IAsyncEventBus):
    _transport: IAsyncEventTransport
    _adapters: dict[type, TypeAdapter[AbstractIntegrationEvent]]

    def __init__(self, transport: IAsyncEventTransport) -> None:
        self._transport = transport
        self._adapters = {}

    async def send(self, event: AbstractIntegrationEvent) -> None:
        event_type = type(event)
        if event_type not in self._adapters:
            self._adapters[event_type] = TypeAdapter(event_type)
        adapter = self._adapters[event_type]
        await self._transport.send(event.event_name, adapter.dump_json(event))



"""Event handler registration post-processor."""

from inspect import getmembers, iscoroutinefunction, ismethod
from logging import getLogger

from spakky.core.pod.annotations.pod import Pod
from spakky.core.pod.interfaces.aware.container_aware import IContainerAware
from spakky.core.pod.interfaces.container import IContainer
from spakky.core.pod.interfaces.post_processor import IPostProcessor
from spakky.domain.models.event import AbstractDomainEvent, AbstractEvent

from spakky.event.event_consumer import (
    IAsyncEventConsumer,
    IEventConsumer,
)
from spakky.event.stereotype.event_handler import EventHandler, EventRoute

logger = getLogger(__name__)


@Pod()
class EventHandlerRegistrationPostProcessor(IPostProcessor, IContainerAware):
    __container: IContainer

    def set_container(self, container: IContainer) -> None:
        self.__container = container

    def post_process(self, pod: object) -> object:
        pod_type = type(pod)

        if not EventHandler.exists(pod_type):
            return pod

        sync_consumer = self.__container.get(IEventConsumer)
        async_consumer = self.__container.get(IAsyncEventConsumer)

        for name, method in getmembers(pod, predicate=ismethod):
            route = EventRoute[AbstractEvent].get_or_none(method)
            if route is None:
                continue
            if not issubclass(route.event_type, AbstractDomainEvent):
                continue

            event_type = route.event_type

            if iscoroutinefunction(method):
                async_consumer.register(event_type, method)
                logger.debug(
                    f"Registered async handler {pod_type.__name__}.{name} "
                    f"for {event_type.__name__}"
                )
            else:
                sync_consumer.register(event_type, method)
                logger.debug(
                    f"Registered sync handler {pod_type.__name__}.{name} "
                    f"for {event_type.__name__}"
                )

        return pod



"""Spakky Event Bus package.

This package provides the default EventBus implementation that delegates
to an EventTransport for actual message delivery.
"""

from spakky.event.bus.transport_event_bus import (
    AsyncDirectEventBus,
    DirectEventBus,
)

__all__ = [
    "AsyncDirectEventBus",
    "DirectEventBus",
]



"""Event dispatcher interfaces for dispatching events to registered handlers.

This module provides unified dispatcher interfaces that handle all event types.
Dispatchers are responsible for delivering events to registered handlers, while
Consumers are responsible for handler registration. These interfaces are combined
in Mediator implementations.
"""

from abc import ABC, abstractmethod

from spakky.domain.models.event import AbstractEvent


class IEventDispatcher(ABC):
    @abstractmethod
    def dispatch(self, event: AbstractEvent) -> None: ...


class IAsyncEventDispatcher(ABC):
    @abstractmethod
    async def dispatch(self, event: AbstractEvent) -> None: ...



"""Spakky Event Mediator package.

This package provides mediator implementations that combine Consumer and Dispatcher
interfaces for in-process event handling.
"""

from spakky.event.mediator.domain_event_mediator import (
    AsyncEventMediator,
    EventMediator,
)

__all__ = [
    "AsyncEventMediator",
    "EventMediator",
]



from abc import ABC, abstractmethod

from spakky.domain.models.event import AbstractEvent, AbstractIntegrationEvent


class IEventPublisher(ABC):
    @abstractmethod
    def publish(self, event: AbstractEvent) -> None: ...


class IAsyncEventPublisher(ABC):
    @abstractmethod
    async def publish(self, event: AbstractEvent) -> None: ...


class IEventBus(ABC):
    @abstractmethod
    def send(self, event: AbstractIntegrationEvent) -> None: ...


class IAsyncEventBus(ABC):
    @abstractmethod
    async def send(self, event: AbstractIntegrationEvent) -> None: ...


class IEventTransport(ABC):
    @abstractmethod
    def send(self, event_name: str, payload: bytes) -> None: ...


class IAsyncEventTransport(ABC):
    @abstractmethod
    async def send(self, event_name: str, payload: bytes) -> None: ...



"""Event mediator implementations.

This module provides in-process mediator implementations that combine Consumer
and Dispatcher interfaces. Mediators manage handler registration and event
dispatching within the same bounded context.
"""

from logging import getLogger
from typing import Any

from spakky.core.pod.annotations.pod import Pod
from spakky.domain.models.event import AbstractEvent

from spakky.event.event_consumer import (
    AsyncEventHandlerCallback,
    EventHandlerCallback,
    IAsyncEventConsumer,
    IEventConsumer,
)
from spakky.event.event_dispatcher import (
    IAsyncEventDispatcher,
    IEventDispatcher,
)

logger = getLogger(__name__)


@Pod()
class EventMediator(IEventConsumer, IEventDispatcher):
    _handlers: dict[type[AbstractEvent], list[EventHandlerCallback[Any]]]

    def __init__(self) -> None:
        self._handlers = {}

    def register(
        self,
        event: type[AbstractEvent],
        handler: EventHandlerCallback[Any],
    ) -> None:
        if event not in self._handlers:
            self._handlers[event] = []
        self._handlers[event].append(handler)
        logger.debug(f"Registered handler for {event.__name__}")

    def dispatch(self, event: AbstractEvent) -> None:
        event_type = type(event)
        handlers = self._handlers.get(event_type, [])

        if not handlers:
            logger.debug(f"No handlers registered for {event_type.__name__}")
            return

        for handler in handlers:
            try:
                handler(event)
            except Exception as e:
                logger.error(
                    f"Handler {handler} failed for {event_type.__name__}: {e}",
                    exc_info=True,
                )


@Pod()
class AsyncEventMediator(IAsyncEventConsumer, IAsyncEventDispatcher):
    _handlers: dict[type[AbstractEvent], list[AsyncEventHandlerCallback[Any]]]

    def __init__(self) -> None:
        self._handlers = {}

    def register(
        self,
        event: type[AbstractEvent],
        handler: AsyncEventHandlerCallback[Any],
    ) -> None:
        if event not in self._handlers:
            self._handlers[event] = []
        self._handlers[event].append(handler)
        logger.debug(f"Registered async handler for {event.__name__}")

    async def dispatch(self, event: AbstractEvent) -> None:
        event_type = type(event)
        handlers = self._handlers.get(event_type, [])

        if not handlers:
            logger.debug(f"No handlers registered for {event_type.__name__}")
            return

        for handler in handlers:
            try:
                await handler(event)
            except Exception as e:
                logger.error(
                    f"Handler {handler} failed for {event_type.__name__}: {e}",
                    exc_info=True,
                )
